from django.contrib.auth.models import User
from allauth.socialaccount.models import SocialAccount, SocialApp, SocialToken
import factory

from ..models import DataportenUser

# See the direct inspiration of these factories here:
# https://factoryboy.readthedocs.io/en/latest/recipes.html#example-django-s-profile


class SocialAccountFactory(factory.django.DjangoModelFactory):

    user = factory.SubFactory(
            'dataporten.tests.factories.DataportenUserFactory',
            social_account=None,
    )
    social_token = factory.RelatedFactory(
            'dataporten.tests.factories.SocialTokenFactory',
            'account',
    )
    provider = 'dataporten'

    class Meta:
        model = SocialAccount


class SocialAppFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = SocialApp


class SocialTokenFactory(factory.django.DjangoModelFactory):
    token = 'dummy_token'
    account = factory.SubFactory(SocialAccountFactory, social_token=None)
    app = factory.SubFactory(SocialAppFactory)

    class Meta:
        model = SocialToken


class UserFactory(factory.django.DjangoModelFactory):
    username = 'username'

    class Meta:
        model = User


class DataportenUserFactory(factory.django.DjangoModelFactory):
    username = 'username'
    social_account = factory.RelatedFactory(SocialAccountFactory, 'user')

    class Meta:
        model = DataportenUser



from collections import defaultdict
from typing import DefaultDict, List, Type, Tuple, Dict

from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

from allauth.socialaccount.models import SocialToken

from .api import usergroups
from .parsers import (
    BaseGroup,
    Course,
    PARSERS,
    Semester,
    group_factory,
)

class DataportenGroupManager:
    """
    A manager for all the different types of groups which dataporten provides.
    It gives access to the following properties:
        - courses
        - active_courses
        - inactive_courses
        - generic_groups
    """
    def __init__(self, token: str) -> None:
        # Fetch usergroups and insert into dictionary based on dataporten unique
        # id
        groups_json = usergroups(token)
        self.groups = {
            group.uid: group
            for group
            in group_factory(*groups_json)
        }

        # Make each group type NAME a direct property of the object itself,
        # each property containing a dictionary keyed on dataporten unique ids.
        for parser in PARSERS:
            setattr(self, parser.NAME, {})

        # Sort all groups into a dictionary keyed on the group type, uniquely
        # represented by the group type's NAME attribute.
        for uid, group in self.groups.items():
            getattr(self, group.NAME)[uid] = group

        self.courses = CourseManager(self.courses)  # type: ignore

    def is_member_of(self, uid: str, active: bool = True) -> bool:
        """
        Return True if user is a member of the given group UID.

        :param uid: The group uid identifying the dataporten group.
          For example: 'fc:org:ntnu.no:unit:167500', which represents
          Orakeltjenesten, NTNU-IT.
        :param active: If True, an *active* membership is required. If False,
          True is returned if the user has been a member of the group at *any*
          point in time, and the user *might* not have active group membership
          anymore.
        """
        group = self.groups.get(uid)

        if not group:
            return False
        elif group and not active:
            return True
        else:
            return bool(group.membership)


class CourseManager:
    def __init__(self, courses: Dict[str, Course]) -> None:
        self.all = {course.code: course for course in courses.values()}
        self.semesters_ago: List[Tuple[int, str]] = []
        now = Semester.now()

        for course in courses.values():
            ago = now - course.semester
            self.semesters_ago.append((ago, course.code,))

    @property
    def active(self) -> List[str]:
        return [code for ago, code in self.semesters_ago if ago <= 0]

    @property
    def finished(self) -> List[str]:
        return [code for ago, code in self.semesters_ago if ago > 0]

    def less_semesters_ago(self, than) -> List[str]:
        return [code for ago, code in self.semesters_ago if ago < than]


class DataportenUser(User):
    """
    Adds a dataporten property to the user, which points to a
    DataportenGroupManager instance. This property is cached
    for the lifetime of the user instance, as the manager makes
    API requests to dataporten
    """

    class Meta:
        proxy = True

    @cached_property
    def token(self) -> str:
        try:
            token_func_path = settings.DATAPORTEN_TOKEN_FUNCTION
            token_func = import_string(token_func_path)
            return token_func(self)
        except ModuleNotFoundError:
            raise ImproperlyConfigured(
                f'Could not import DATAPORTEN_TOKEN_FUNCTION with value '
                f'{token_func_path}',
            )
        except AttributeError:
            raise ImproperlyConfigured(
                'You need to define DATAPORTEN_TOKEN_FUNCTION in your '
                'settings.py',
            )


    @cached_property
    def dataporten(self):
        return DataportenGroupManager(self.token)

    @staticmethod
    def valid_request(request: HttpRequest) -> bool:
        if hasattr(request, 'user') and request.user.is_authenticated:
            return SocialToken.objects.filter(
                account__user=request.user,
                account__provider='dataporten',
            ).exists()
        else:
            return False



""" Dataporten test settings and fixtures """

from django.contrib.auth.models import AnonymousUser, User
from freezegun import freeze_time
import pytest
import responses

from .factories import DataportenUserFactory, UserFactory
from .utils import mock_usergroups_request
from ..models import DataportenGroupManager
from ..parsers import Course


@pytest.fixture
@responses.activate
def dataporten():
    group_json = mock_usergroups_request()
    return DataportenGroupManager('dummy_token')


@pytest.fixture
@pytest.mark.django_db
def dataporten_user():
    return DataportenUserFactory()


@pytest.fixture
@pytest.mark.django_db
def user_with_dataporten_token():
    """ A User which has not been swapped out with the proxy model """
    dp_user = DataportenUserFactory()
    dp_user.__class__ = User
    return dp_user


# Examples of json structures received from Dataporten representing groups
@pytest.fixture
def study_program_json():
    return {
        "displayName": "Fysikk og matematikk - masterstudium (5-årig)",
        "membership": {
            "basic": "member",
            "displayName": "Student",
            "active": True,
            "fsroles": [
                "STUDENT"
            ]
        },
        "parent": "fc:org:ntnu.no",
        "url": "http://www.ntnu.no/studier/mtfyma",
        "id": "fc:fs:fs:prg:ntnu.no:MTFYMA",
        "type": "fc:fs:prg",
    }


@pytest.fixture
def course_json():
    return {
        "displayName": "Examen philosophicum for naturvitenskap og teknologi",
        "id": "fc:fs:fs:emne:ntnu.no:EXPH0004:1",
        "parent": "fc:org:ntnu.no",
        "type": "fc:fs:emne",
        "membership": {
            "displayName": "Student",
            "notAfter": "2014-12-14T23:00:00Z",
            "active": True,
            "fsroles": [
                "STUDENT"
            ],
            "subjectRelations": "undervisning",
            "basic": "member"
        },
        "url": "http://www.ntnu.no/exphil"
    }


@pytest.fixture
def main_profile_json():
    return {
        "parent": "fc:org:ntnu.no",
        "displayName": "Industriell matematikk",
        "type": "fc:fs:str",
        "membership": {
            "displayName": "Student",
            "fsroles": [
                "STUDENT"
            ],
            "basic": "member",
            "active": True
        },
        "id": "fc:fs:fs:str:ntnu.no:MTFYMA-IM"
    }


@pytest.fixture
def organisation_unit_json():
    return {
        "displayName": "Seksjon for IT-brukerst\u00f8tte",
        "id": "fc:org:ntnu.no:unit:167500",
        "public": True,
        "parent": "fc:org:ntnu.no",
        "type": "fc:orgunit",
        "membership": {
            "primaryOrgUnit": True,
            "basic": "member"
        }
    }


@pytest.fixture
def finished_course():
    return Course(
        {
            "displayName": "Examen philosophicum for naturvitenskap og teknologi",
            "membership": {
                "notAfter": "2014-12-14T23:00:00Z",
                "active": True,
                "subjectRelations": "undervisning",
                "basic": "member",
                "fsroles": [
                    "STUDENT"
                ],
                "displayName": "Student"
            },
            "parent": "fc:org:ntnu.no",
            "url": "http://www.ntnu.no/exphil",
            "id": "fc:fs:fs:emne:ntnu.no:EXPH0004:1",
            "type": "fc:fs:emne",
        }
    )


@freeze_time('2017-01-01')
@pytest.fixture
def non_finished_course():
    return Course(
        {
            "displayName": "Algebra ",
            "membership": {
                "basic": "member",
                "displayName": "Student",
                "active": True,
                "fsroles": [
                    "STUDENT"
                ]
            },
            "parent": "fc:org:ntnu.no",
            "url": "http://wiki.math.ntnu.no/tma4150",
            "id": "fc:fs:fs:emne:ntnu.no:TMA4150:1",
            "type": "fc:fs:emne"
        }
    )


@pytest.fixture
def ongoing_course():
    return Course(
        {
            "displayName": "Optimering I",
            "membership": {
                "notAfter": "2017-08-14T22:00:00Z",
                "active": True,
                "subjectRelations": "undervisning",
                "basic": "member",
                "fsroles": [
                    "STUDENT"
                ],
                "displayName": "Student"
            },
            "parent": "fc:org:ntnu.no",
            "url": "http://wiki.math.ntnu.no/tma4180",
            "id": "fc:fs:fs:emne:ntnu.no:TMA4180:1",
            "type": "fc:fs:emne"
        }
    )


@pytest.fixture
def course_last_semester():
    return Course(
        {
            "displayName": "Line\u00e6re metoder",
            "id": "fc:fs:fs:emne:ntnu.no:TMA4145:1",
            "parent": "fc:org:ntnu.no",
            "type": "fc:fs:emne",
            "membership": {
                "displayName": "Student",
                "notAfter": "2016-12-14T23:00:00Z",
                "active": True,
                "fsroles": [
                    "STUDENT"
                ],
                "subjectRelations": "undervisning",
                "basic": "member"
            },
            "url": "http://wiki.math.ntnu.no/tma4145"
        }
    )


@pytest.fixture
def organization_json():
    return {
        "id": "fc:org:ntnu.no",
        "displayName": "NORGES TEKNISK-NATURVITENSKAPELIGE UNIVERSITET NTNU",
        "street": "N-7491",
        "eduOrgLegalName": "NORGES TEKNISK-NATURVITENSKAPELIGE UNIVERSITET NTNU",
        "eduOrgHomePageURI": "http://www.ntnu.no/",
        "norEduOrgNIN": "N1111111111",
        "facsimileTelephoneNumber": "+47 12345678",
        "norEduOrgAcronym": "NTNU",
        "labeledURI": "http://www.ntnu.no/",
        "public": True,
        "mail": "orakel@ntnu.no",
        "orgType": ["higher_education"],
        "membership": {
            "title": ["fast ansatt"],
            "affiliation": ["employee", "member", "affiliate", "student"],
            "primaryAffiliation": "employee",
            "basic": "admin",
            "displayName": "Ansatt",
        },
        "postalAddress": "N-7491 Trondheim $ NORWAY",
        "norEduOrgUniqueIdentifier": "194000000",
        "postalCode": "N-7491",
        "type": "fc:org",
        "telephoneNumber": "+47 73595000",
        "l": "Trondheim",
    }


@pytest.fixture
def membership_json(organization_json):
    return organization_json["membership"]



"""
Tools for gathering userinfo and groups for a specific dataporten user
"""
from typing import Any, Dict, List, Optional

import requests
from mypy_extensions import TypedDict


# All the required fields of the Dataporten
# JSON representation of groups memberships
class MembershipJSONBase(TypedDict):
    basic: str  # NOQA
    displayName: str
    fsroles: List[str]


# Optional fields
class MembershipJSON(MembershipJSONBase, total=False):
    active: bool
    notAfter: str
    subjectRelations: Optional[str]


# All the required fields of the Dataporten
# JSON representation of groups memberships
class GroupJSONBase(TypedDict):
    displayName: str
    parent: str
    url: str
    id: str
    type: str


# Optional fields
class GroupJSON(GroupJSONBase, total=False):
    membership: MembershipJSON


def userinfo(token: str) -> Dict[str, Any]:
    USERINFO_URL = 'https://auth.dataporten.no/userinfo'

    # The athentication header
    headers = {'Authorization': 'Bearer ' + token}

    # Userinfo endpoint, for documentation see:
    # https://docs.dataporten.no/docs/oauth-authentication/
    userinfo_response = requests.get(
        USERINFO_URL,
        headers=headers,
    )
    # Raise exception for 4xx and 5xx response codes
    userinfo_response.raise_for_status()

    # The endpoint returns json-data and it needs to be decoded
    return userinfo_response.json()['user']


def usergroups(token: str) -> List[GroupJSON]:
    GROUPS_URL = 'https://groups-api.dataporten.no/groups/'

    # The athentication header
    headers = {'Authorization': 'Bearer ' + token}

    # Groups endpoint, for documentation see:
    # https://docs.dataporten.no/docs/groups/
    groups_data = requests.get(
        GROUPS_URL + 'me/groups',
        headers=headers,
    )
    # Raise exception for 4xx and 5xx response codes
    groups_data.raise_for_status()

    # The endpoint returns json-data and it needs to be decoded
    return groups_data.json()



import json
from typing import Any, Dict, List

from django.contrib.auth.models import User

import responses


def mock_userinfo_request() -> Dict[str, Any]:
    """
    Mocks the next request to the userinfo dataporten endpoint.
    Returns the mocked response, but already json decoded
    """
    with open('dataporten/tests/userinfo.json') as userinfo_file:
        userinfo_dump = json.load(userinfo_file)

    responses.add(
        responses.GET,
        'https://auth.dataporten.no/userinfo',
        json=userinfo_dump,
        status=200,
    )

    return userinfo_dump


def mock_usergroups_request() -> List[Dict[Any, Any]]:
    """
    Mocks the next request to the groups dataporten endpoint.
    Returns the mocked response, but already json decoded
    """
    with open('dataporten/tests/groups.json') as groups_file:
        groups_dump = json.load(groups_file)

    responses.add(
        responses.GET,
        'https://groups-api.dataporten.no/groups/me/groups',
        json=groups_dump,
        status=200,
    )

    return groups_dump


def dummy_token_function(user: User) -> str:
    return 'dummy_token'



from .settings import *

SECRET_KEY = 'dummy'

# Using an in-memory sqlite3 database for faster testing
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

INSTALLED_APPS = (
    # Standard Django apps
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',

    # Dataporten app
    'dataporten',

    # oAuth2 framework
    'allauth',
)

# Cache requests to the dataporten API
DATAPORTEN_CACHE_REQUESTS = True

# Where to save the sqlite3 cache backend
DATAPORTEN_CACHE_PATH = 'tmp/'

# Dummy token function
DATAPORTEN_TOKEN_FUNCTION = 'dataporten.tests.utils.dummy_token_function'



import json

import responses
from django.test import TestCase

from ..api import usergroups, userinfo
from .utils import mock_usergroups_request, mock_userinfo_request


class TestTypes(TestCase):
    """ Should hit type definitions in order to catch syntax errors """
    from ..api import GroupJSON


class TestUserInfo(TestCase):
    """ Test dataporten userinfo endpoint """

    @responses.activate
    def test_valid_case(self):
        userinfo_dict = mock_userinfo_request()
        userinfo_return = userinfo('testac')
        self.assertEqual(
            userinfo_return,
            userinfo_dict['user'],
        )


class TestGroups(TestCase):
    """ Test dataporten groups endpoint """

    @responses.activate
    def test_valid_case(self):
        groups_dict = mock_usergroups_request()
        userinfo_return = usergroups('testac')
        self.assertEqual(
            userinfo_return,
            groups_dict,
        )



from django.contrib.auth.models import AnonymousUser

import pytest
import responses
from django.core.exceptions import ImproperlyConfigured
from freezegun import freeze_time

from .utils import mock_usergroups_request
from ..models import (
    DataportenGroupManager,
    CourseManager,
    DataportenUser,
)


@freeze_time('2016-01-01')
class TestDataportenGroupManager:
    def test_dataporten_courses(self, dataporten):
        # End time date in past
        assert 'EXPH0004' in dataporten.courses.finished
        assert 'EXPH0004' in dataporten.courses.all
        assert 'EXPH0004' not in dataporten.courses.active

        # Alternative way to check membership, where active gives different
        # results since this is
        assert dataporten.is_member_of(
            uid='fc:fs:fs:emne:ntnu.no:EXPH0004:1',
            active=False,
        )
        assert not dataporten.is_member_of(
            uid='fc:fs:fs:emne:ntnu.no:EXPH0004:1',
            active=True,
        )

        # End time date in future
        assert 'TMA4150' not in dataporten.courses.finished
        assert 'TMA4150' in dataporten.courses.all
        assert 'TMA4150' in dataporten.courses.active
        assert dataporten.is_member_of(
            uid='fc:fs:fs:emne:ntnu.no:TMA4150:1',
            active=False,
        )
        assert dataporten.is_member_of(
            uid='fc:fs:fs:emne:ntnu.no:TMA4150:1',
            active=True,
        )


        # No end time, implied future date
        assert 'TMA4180' not in dataporten.courses.finished
        assert 'TMA4180' in dataporten.courses.all
        assert 'TMA4180' in dataporten.courses.active
        assert dataporten.is_member_of(
            uid='fc:fs:fs:emne:ntnu.no:TMA4180:1',
            active=False,
        )
        assert dataporten.is_member_of(
            uid='fc:fs:fs:emne:ntnu.no:TMA4180:1',
            active=True,
        )

    def test_dataporten_study_program(self, dataporten):
        assert dataporten.\
            study_programs['fc:fs:fs:prg:ntnu.no:MTFYMA'].\
            code == 'MTFYMA'

    def test_dataporten_main_profile(self, dataporten):
        assert dataporten.\
            main_profiles['fc:fs:fs:str:ntnu.no:MTFYMA-IM'].\
            code == 'MTFYMA-IM'

    def test_dataporten_organisation_units(self, dataporten):
        assert dataporten.\
            organisation_units['fc:org:ntnu.no:unit:167500'].\
            uid == 'fc:org:ntnu.no:unit:167500'

        assert dataporten.\
            organisation_units['fc:org:ntnu.no:unit:660000'].\
            uid == 'fc:org:ntnu.no:unit:660000'


@freeze_time('2017-01-01')
class TestCourseManager:
    def test_less_semesters_ago(self, finished_course, course_last_semester, ongoing_course):
        _courses = [
            finished_course,
            course_last_semester,
            ongoing_course,
        ]
        courses = {course.uid: course for course in _courses}
        course_manager = CourseManager(courses)

        assert course_manager.less_semesters_ago(than=1) \
               == [ongoing_course.code]

        assert course_manager.less_semesters_ago(than=2) \
               == [course_last_semester.code, ongoing_course.code]

        assert course_manager.less_semesters_ago(than=20) \
               == [finished_course.code, course_last_semester.code, ongoing_course.code]

    @freeze_time('1900-01-01')
    def test_semester_in_future(self, ongoing_course):
        """
        The Course manager needs to behave correctly when it is presented with a course
        which ends in the next semester.
        """
        course_manager = CourseManager({ongoing_course.uid: ongoing_course})
        assert ongoing_course.code in course_manager.active


def token_import_function(user):
    return 'dummy_token_from_import_function'


class TestTokenImport:
    @pytest.mark.django_db
    def test_token_not_configured(self, settings, dataporten_user):
        try:
            del settings.DATAPORTEN_TOKEN_FUNCTION
        except AttributeError:
            pass

        with pytest.raises(ImproperlyConfigured) as excinfo:
            dataporten_user.token
        assert str(excinfo.value) == \
               'You need to define DATAPORTEN_TOKEN_FUNCTION in your ' \
               'settings.py'

    @pytest.mark.django_db
    def test_token_unimportable_path(self, settings, dataporten_user):
        settings.DATAPORTEN_TOKEN_FUNCTION = 'highly.unlikely.to.exist'

        with pytest.raises(ImproperlyConfigured) as excinfo:
            dataporten_user.token
        assert str(excinfo.value) == \
               'Could not import DATAPORTEN_TOKEN_FUNCTION with value ' \
               'highly.unlikely.to.exist'

    @pytest.mark.django_db
    def test_token_properly_configured(self, settings, dataporten_user):
        settings.DATAPORTEN_TOKEN_FUNCTION = \
            'dataporten.tests.test_models.token_import_function'

        assert dataporten_user.token == 'dummy_token_from_import_function'


@pytest.mark.django_db
@responses.activate
def test_dataporten_user_groups(dataporten_user):
    group_json = mock_usergroups_request()
    groups = dataporten_user.dataporten
    assert isinstance(groups, DataportenGroupManager)


@pytest.mark.django_db
def test_dataporten_user_validation(rf, django_user_model, dataporten_user):
    # pytest-django provides a RequestFactory fixture named rf
    request = rf.get('')

    request.user = AnonymousUser()
    assert DataportenUser.valid_request(request) == False

    request.user = django_user_model()
    assert DataportenUser.valid_request(request) == False

    request.user = dataporten_user
    assert DataportenUser.valid_request(request) == True



import datetime
from typing import Dict, Generator, Iterable, List, Optional, Tuple

from .api import GroupJSON, MembershipJSON


def datetime_from(json_string: str) -> datetime.datetime:
    return datetime.datetime.strptime(
        json_string,
        '%Y-%m-%dT%H:%M:%SZ',
    )


def group_type(group: GroupJSON) -> str:
    return group['type'].split(':')[-1]


def group_factory(*groups: Iterable[GroupJSON]) -> \
                                       Generator['BaseGroup', None, None]:
    """
    Given a JSON Group structure, this function returns the most specific
    object type, given the input
    """

    for group in groups:
        for parser in PARSERS:
            if parser.valid(group):
                yield parser(group)
                break
        else:
            # In case somebody removed Group from PARSERS, this is a fall back group
            # which will accept all forms of dataporten groups
            yield Group(group)



class BaseGroup:
    """
    A basic class for behaviour present in all Dataporten groups,
    intended for subclassing
    """

    DATAPORTEN_TYPE: Optional[str]
    NAME: str

    def __init__(self, group: GroupJSON) -> None:
        if not self.valid(group):
            raise TypeError('Invalid Group JSON structure')

        self.name = group.get('displayName', 'No display name').strip()
        self.url = group.get('url', '')
        self.uid = group['id']
        self.group_type = group_type(group)

        if 'membership' in group:
            self.membership = Membership(group['membership'])
        else:
            self.membership = None

    @classmethod
    def valid(cls, group: GroupJSON) -> bool:
        """ Subclasses can define their corresponding Dataporten type """
        return cls.DATAPORTEN_TYPE == group_type(group)


class Group(BaseGroup):
    """ A fallback Group type """

    NAME = 'generic_groups'

    @classmethod
    def valid(cls, group: GroupJSON) -> bool:
        # No restrictions are imposed on the JSON for the fallback group as
        # long as the JSON passes through the __init__ method without
        # throwing exceptions
        return True


class Membership:
    def __init__(self, membership: MembershipJSON) -> None:
        self.json = membership
        self.primary_affiliation = membership.get('primaryAffiliation')
        self.affiliations = membership.get('affiliation', [])

        if 'active' in membership:
            self.active = membership['active']
        else:
            # If there is no 'active' key, then the membership
            # is implicitly active (it seems)
            self.active = True

        if 'notAfter' in membership:
            self.end_time = datetime_from(membership['notAfter'])
            self.semester = Semester(self.end_time)
        else:
            self.end_time = None

    def __bool__(self) -> bool:
        """
        The Membership object is truthy if the membership is not only active,
        but also current (old memberships stay active)
        """
        if not self.active:
            return False
        elif self.end_time is None:
            # If there is no end date, we can assume it to be true
            # This is often the case when there has not been set an examination
            # date yet
            return True
        else:
            return datetime.datetime.now() < self.end_time

    def __str__(self) -> str:
        """
        Return string representation of Membership object.

        The `displayName` property is used for this purpose, and if it does
        not exist, "Ukjent" is returned.
        """
        return self.json.get('displayName', 'Ukjent')


def next_holiday() -> datetime.datetime:
    now = datetime.datetime.now()
    if 1 <= now.month <= 6:
        return datetime.datetime(year=now.year, month=6, day=10)
    else:
        return datetime.datetime(year=now.year, month=12, day=22)


class Course(BaseGroup):
    DATAPORTEN_TYPE = 'emne'
    NAME = 'courses'

    def __init__(self, group: GroupJSON) -> None:
        super().__init__(group)
        self.code = group['id'].split(':')[-2]

        if self.membership is not None and self.membership.end_time:
            self.semester = Semester(self.membership.end_time)
        else:
            self.semester = Semester(next_holiday())

    CourseDict = Dict[str, 'Course']

    @classmethod
    def split_on_membership(
            cls,
            courses: List['Course'],
    ) -> Tuple[CourseDict, CourseDict]:

        """
        Return two dictionaries in a tuple (active, inactive,)
        of the form Dict[course_code] = course_object
        """

        active, inactive = {}, {}
        for course in courses:
            if course.membership:
                active[course.code] = course
            else:
                inactive[course.code] = course

        return active, inactive

class StudyProgram(BaseGroup):
    DATAPORTEN_TYPE = 'prg'
    NAME = 'study_programs'

    def __init__(self, group: GroupJSON) -> None:
        super().__init__(group)
        self.code = group['id'].split(':')[-1]


class MainProfile(BaseGroup):
    DATAPORTEN_TYPE = 'str'
    NAME = 'main_profiles'

    def __init__(self, group: GroupJSON) -> None:
        super().__init__(group)
        self.code = group['id'].split(':')[-1]


class OrganisationUnit(BaseGroup):
    DATAPORTEN_TYPE = 'orgunit'
    NAME = 'organisation_units'


class Semester:
    SPRING = 0
    AUTUMN = 1

    def __init__(self, dt: datetime.datetime) -> None:
        self.year = dt.year
        self.season = self.determine_season(dt.month)

    @classmethod
    def determine_season(cls, month: int) -> int:
        if 1 <= month <= 6:
            return cls.SPRING
        else:
            return cls.AUTUMN

    @classmethod
    def now(cls):
        return cls(datetime.datetime.now())

    def __sub__(self, other: 'Semester') -> int:
        return 2 * (self.year - other.year) + (self.season - other.season)


PARSERS = [
    Course,
    StudyProgram,
    MainProfile,
    OrganisationUnit,
    Group,
]



from datetime import datetime

from django.test import TestCase

from freezegun import freeze_time

import pytest

from ..parsers import (
        Course,
        Group,
        group_factory,
        MainProfile,
        Membership,
        OrganisationUnit,
        Semester,
        StudyProgram,
        datetime_from,
)


class TestDatetimeFrom(TestCase):
    def test_basic_correctness(self):
        dt = datetime_from('2017-08-14T22:00:01Z')
        self.assertEqual(
            [dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second],
            [2017, 8, 14, 22, 0, 1],
        )


class TestGroupFactory:
    def test_study_program_factory(self, study_program_json):
        study_program = next(group_factory(study_program_json))
        assert type(study_program) is StudyProgram

    def test_course_factory(self, course_json):
        course = next(group_factory(course_json))
        assert type(course) is Course

    def test_main_profile_factory(self, main_profile_json):
        main_profile = next(group_factory(main_profile_json))
        assert type(main_profile) is MainProfile

    def test_group_factory_given_iterable_argument(
            self,
            study_program_json,
            course_json,
    ):
        groups = group_factory(study_program_json, course_json)
        assert len(list(groups)) == 2


class TestGroup:
    def test_properties_present(self, study_program_json):
        group_example = Group(study_program_json)
        assert group_example.name == 'Fysikk og matematikk - masterstudium (5-årig)'
        assert group_example.url == 'http://www.ntnu.no/studier/mtfyma'
        assert group_example.group_type == 'prg'

    def test_active_membership(self, study_program_json):
        group_example = Group(study_program_json)
        assert group_example.membership


class TestMembership():
    def test_perpetual_membership(self):
        unending_membership = Membership(
            {
                "basic": "member",
                "displayName": "Student",
                "active": True,
                "fsroles": [
                    "STUDENT"
                ]
            }
        )
        assert bool(unending_membership) is True

    def test_inactive_membership(self):
        inactive_membership = Membership(
            {
                "basic": "member",
                "displayName": "Student",
                "active": False,
                "fsroles": [
                    "STUDENT"
                ]
            }
        )
        assert bool(inactive_membership) is False

    @freeze_time('2017-08-13')
    def test_limited_membership(self):
        limited_membership = Membership(
            {
                "notAfter": "2017-08-14T22:00:00Z",
                "active": True,
                "subjectRelations": "undervisning",
                "basic": "member",
                "fsroles": [
                    "STUDENT"
                ],
                "displayName": "Student"
            }
        )
        assert bool(limited_membership) is True

    @freeze_time('2017-08-15')
    def test_expired_membership(self):
        limited_membership = Membership(
            {
                "notAfter": "2017-08-14T22:00:00Z",
                "active": True,
                "subjectRelations": "undervisning",
                "basic": "member",
                "fsroles": [
                    "STUDENT"
                ],
                "displayName": "Student"
            }
        )
        assert bool(limited_membership) is False

    def test_retrieval_of_json_from_membership_object(self, membership_json):
        """Original JSON should be stored on the membership object."""
        membership = Membership(membership_json)
        assert membership.json == membership_json


    def test_string_representation_of_membership(self, membership_json):
        """The displayName property should be used for str representation."""
        # With a displayName attribute
        membership = Membership(membership_json)
        assert str(membership) == 'Ansatt'

        # Without a displayName attribute
        del membership_json['displayName']
        membership = Membership(membership_json)
        assert str(membership) == 'Ukjent'

    def test_primary_affiliation_property(self, membership_json):
        """Membership affiliations should be retrievable."""
        membership = Membership(membership_json)
        assert membership.primary_affiliation == 'employee'
        assert membership.affiliations == [
            'employee',
            'member',
            'affiliate',
            'student',
        ]


@freeze_time('2017-01-01')
class TestCourse:
    def test_course_code(self, finished_course):
        assert finished_course.code == 'EXPH0004'

    def test_finished_course(self, finished_course):
        assert not finished_course.membership
        assert finished_course.semester.year == 2014

    def test_ongoing_course_with_end_time(self, ongoing_course):
        assert ongoing_course.membership
        assert ongoing_course.semester.year == 2017

    @pytest.mark.skip(reason='Flaky for some versions. TODO.')
    def test_ongoing_course_without_end_time(self, non_finished_course):
        assert non_finished_course.membership
        assert non_finished_course.semester.year == 2019

    def test_split_on_membership(self, finished_course, non_finished_course, ongoing_course):
        courses = [
            finished_course,
            non_finished_course,
            ongoing_course,
        ]
        active, inactive = Course.split_on_membership(courses)

        assert finished_course.code in inactive.keys()
        assert finished_course in inactive.values()

        assert non_finished_course.code in active.keys()
        assert non_finished_course in active.values()

        assert ongoing_course.code in active.keys()
        assert ongoing_course in active.values()


class TestStudyProgram:
    def test_study_program_basic_properties(self, study_program_json):
        study_program = StudyProgram(study_program_json)
        assert study_program.code == 'MTFYMA'


class TestMainProfile:
    def test_main_profile_basic_properties(self, main_profile_json):
        main_profile = MainProfile(main_profile_json)
        assert main_profile.code == 'MTFYMA-IM'


class TestOrganisationUnit:
    def test_organisation_unit_basic_properties(self, organisation_unit_json):
        organisation_unit = OrganisationUnit(organisation_unit_json)
        assert organisation_unit.uid == 'fc:org:ntnu.no:unit:167500'


@freeze_time('2017-08-27')
class TestSemester(TestCase):
    def setUp(self):
        autumn_semester_date = datetime_from('2016-09-14T22:00:00Z')
        spring_semester_date = datetime_from('2016-04-04T22:00:00Z')

        self.autumn_semester = Semester(autumn_semester_date)
        self.spring_semester = Semester(spring_semester_date)
        self.present_semester = Semester.now()

    def test_year_of_semester(self):
        self.assertEqual(self.autumn_semester.year, 2016)
        self.assertEqual(self.present_semester.year, 2017)

    def test_semester_season(self):
        self.assertEqual(self.autumn_semester.season, Semester.AUTUMN)
        self.assertEqual(self.spring_semester.season, Semester.SPRING)

    def test_subtracting_semesters(self):
        same_semester_diff = self.present_semester - self.present_semester
        same_season_diff = self.present_semester - self.autumn_semester
        negative_diff = self.autumn_semester - self.present_semester
        different_season_diff = self.autumn_semester - self.spring_semester
        different_season_negative_diff = self.spring_semester - self.present_semester

        self.assertEqual(
            [same_semester_diff, same_season_diff, negative_diff, different_season_diff, different_season_negative_diff],
            [0, 2, -2, 1, -3],
        )



from unittest.mock import Mock

from django.contrib.auth.models import AnonymousUser, User

import pytest

from ..middleware import DataportenGroupsMiddleware
from ..models import DataportenUser


def test_dataporten_middleware_with_anonymous_user(rf, monkeypatch):
    request = rf.get('')
    dpm = DataportenGroupsMiddleware(Mock())

    request.user = AnonymousUser()
    dpm(request)
    assert type(request.user) is AnonymousUser


@pytest.mark.django_db
def test_dataporten_middleware_with_plain_django_user(
    rf,
    django_user_model,
    monkeypatch,
):
    request = rf.get('')
    dpm = DataportenGroupsMiddleware(Mock())

    request.user = django_user_model()
    dpm(request)
    assert type(request.user) is User


@pytest.mark.django_db
def test_dataporten_middleware_with_user_with_dataporten_credentials(
    rf,
    user_with_dataporten_token,
    monkeypatch,
):
    request = rf.get('')
    dpm = DataportenGroupsMiddleware(Mock())

    request.user = user_with_dataporten_token
    dpm(request)
    assert type(request.user) is DataportenUser



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-dataporten',
    version='0.4.0',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='A simple Django app to fetch and parse data from Dataporten.',
    long_description=README,
    url='https://github.com/JakobGM/django-dataporten',
    author='Jakob Gerhard Martinussen',
    author_email='jakobgm@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
    ],
    install_requires=[
        'django',
        'pip-tools',
        'pytest',
        'pytest-django',
        'requests',
        'requests-cache',
        'responses',
        'mypy',
        'mypy_extensions',
        'freezegun',
        'factory_boy',
        'django-allauth'
    ]
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-07-27 16:58
from __future__ import unicode_literals

import django.contrib.auth.models
from django.db import migrations


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataportenUser',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('auth.user',),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
    ]



import logging

import requests
import requests_cache

from django.conf import settings
from django.http import HttpResponse, HttpRequest
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

from .api import usergroups
from .models import DataportenUser

# Cache requests for 15 minutes
if settings.DATAPORTEN_CACHE_REQUESTS:
    requests_cache.install_cache(
        settings.DATAPORTEN_CACHE_PATH + 'dataporten_cache',
        backend='sqlite',
        expire_after=900,
        allowable_codes=(200,),
        include_get_headers=True,
    )


class DataportenGroupsMiddleware(MiddlewareMixin):
    def process_request(self, request: HttpRequest):
        if DataportenUser.valid_request(request):
            request.user.__class__ = DataportenUser
